from py_order_utils.builders import LimitOrderBuilder, MarketOrderBuilder
from py_order_utils.model import EOA, LimitOrderData, MarketOrderData
from py_order_utils.config import get_contract_config

from .helpers import round_down, to_token_decimals
from .constants import BUY

from ..signer import Signer
from ..clob_types import LimitOrderArgs, MarketOrderArgs


class OrderBuilder:
    
    def __init__(self, signer: Signer, sig_type=None, funder=None):
        self.signer = signer
        
        # Signature type used sign Limit and Market orders, defaults to EOA type
        self.sig_type = sig_type if sig_type is not None else EOA

        # Address which holds funds to be used.
        # Used for Polymarket proxy wallets and other smart contract wallets
        # Defaults to the address of the signer
        self.funder = funder if funder is not None else self.signer.address
        self.contract_config = self._get_contract_config(self.signer.get_chain_id())
        self.limit_order_builder = LimitOrderBuilder(self.contract_config.exchange, self.signer.chain_id, self.signer)
        self.market_order_builder = MarketOrderBuilder(self.contract_config.exchange, self.signer.chain_id, self.signer)

    def _get_contract_config(self, chain_id: int):
        return get_contract_config(chain_id)

    def _create_limit_order_builder(self):
        return LimitOrderBuilder(self.contract_config.exchange, self.signer.chain_id, self.signer)


    def create_limit_order(self, order_args: LimitOrderArgs):
        """
        Creates and signs a limit order
        """
        if order_args.side == BUY:
            maker_asset = self.contract_config.get_collateral()
            taker_asset = self.contract_config.get_conditional()
            maker_asset_id = None
            taker_asset_id = int(order_args.token_id)

            size_normalized = round_down(order_args.size, 2)
            maker_amount = to_token_decimals(round_down(order_args.price * size_normalized, 2))
            taker_amount = to_token_decimals(size_normalized)
        else:
            maker_asset = self.contract_config.get_conditional()
            taker_asset = self.contract_config.get_collateral()
            maker_asset_id = int(order_args.token_id)
            taker_asset_id = None

            size_normalized = round_down(order_args.size, 2)
            maker_amount = to_token_decimals(size_normalized)
            taker_amount = to_token_decimals(round_down(order_args.price * size_normalized, 2))

        data = LimitOrderData(
                exchange_address=self.contract_config.get_exchange(),
                maker_asset_address=maker_asset,
                maker_asset_id=maker_asset_id,
                taker_asset_address=taker_asset,
                taker_asset_id=taker_asset_id,
                maker_address=self.funder,
                maker_amount=maker_amount,
                taker_amount=taker_amount,
                signer=self.signer.address,
                sig_type=self.sig_type
        )
        
        return self.limit_order_builder.create_limit_order(data)

    def create_market_order(self, order_args: MarketOrderArgs):
        """
        """
        if order_args.side == BUY:
            maker_asset = self.contract_config.get_collateral()
            taker_asset = self.contract_config.get_conditional()
            maker_asset_id = None
            taker_asset_id = int(order_args.token_id)

            size_normalized = round_down(order_args.size, 2)
            maker_amount = to_token_decimals(round_down(order_args.price * size_normalized, 2))
        else:
            maker_asset = self.contract_config.get_conditional()
            taker_asset = self.contract_config.get_collateral()
            maker_asset_id = int(order_args.token_id)
            taker_asset_id = None

            size_normalized = round_down(order_args.size, 2)
            maker_amount = to_token_decimals(size_normalized)

        data = MarketOrderData(
                exchange_address=self.contract_config.get_exchange(),
                maker_asset_address=maker_asset,
                maker_asset_id=maker_asset_id,
                taker_asset_address=taker_asset,
                taker_asset_id=taker_asset_id,
                maker_address=self.funder,
                maker_amount=maker_amount,
                signer=self.signer.address,
                sig_type=self.sig_type
        )        
        return self.market_order_builder.create_market_order(data)



from math import floor

def round_down(x: float, sig_digits: int)-> float:
    return floor(x * (10 ** sig_digits)) / (10 ** sig_digits)

def to_token_decimals(x: float)-> int:
    return int((10 ** 6) * x)



from .clob_types import ApiCreds, RequestArgs
from .signing.hmac import build_hmac_signature
from .signer import Signer
from .signing.eip712 import sign_clob_auth_message
from datetime import datetime

POLY_ADDRESS = "POLY_ADDRESS"
POLY_SIGNATURE = "POLY_SIGNATURE"
POLY_TIMESTAMP = "POLY_TIMESTAMP"
POLY_API_KEY = "POLY_API_KEY"
POLY_PASSPHRASE = "POLY_PASSPHRASE"


def create_level_1_headers(signer: Signer):
    """
    Creates Level 1 Poly headers for a request
    """
    timestamp = int(datetime.now().timestamp())
    signature = sign_clob_auth_message(signer, timestamp)
    headers = {
        POLY_ADDRESS: signer.address,
        POLY_SIGNATURE: signature,
        POLY_TIMESTAMP: str(timestamp),
    }
    return headers



def create_level_2_headers(signer: Signer, creds: ApiCreds, request_args: RequestArgs):
    """
    Creates Level 2 Poly headers for a request
    """
    timestamp = int(datetime.now().timestamp())
    
    hmac_sig = build_hmac_signature(
        creds.api_secret, 
        timestamp, 
        request_args.method, 
        request_args.request_path, 
        request_args.body
    )

    return {
        POLY_ADDRESS: signer.address,
        POLY_SIGNATURE: hmac_sig,
        POLY_TIMESTAMP: str(timestamp),
        POLY_API_KEY: creds.api_key,
        POLY_PASSPHRASE: creds.api_passphrase,
    }




import requests
import json

from ..exceptions import PolyApiException

GET = "GET"
POST = "POST"
DELETE = "DELETE"
PUT = "PUT"

def request(endpoint: str, method:str, headers=None, data=None):
    try:
        resp = requests.request(method=method, url=endpoint, headers=headers, json=data if data else None)
        if resp.status_code != 200:
            raise PolyApiException(resp)
        return resp.json()
    except requests.RequestException:
        raise PolyApiException(error_msg="Request exception!")

def post(endpoint, headers=None, data=None):
    return request(endpoint, POST, headers, data)

def get(endpoint, headers=None, data=None):
    return request(endpoint, GET, headers, data)

def delete(endpoint, headers=None, data=None):
    return request(endpoint, DELETE, headers, data)




from eth_account import Account


class Signer:
    def __init__(self, private_key: str, chain_id: int):
        assert(private_key is not None and chain_id is not None)
        
        self.private_key = private_key
        self.account = Account.from_key(private_key)
        self.chain_id = chain_id

    @property
    def address(self):
        return self.account.address

    def get_chain_id(self):
        return self.chain_id

    def sign(self, message_hash):
        """
        Signs a message hash
        """
        return Account._sign_hash(message_hash, self.private_key).signature.hex()


from eip712_structs import make_domain
from .model import ClobAuth
from ..signer import Signer
from web3 import Web3

CLOB_DOMAIN_NAME = "ClobAuthDomain"
CLOB_VERSION = "1"
MSG_TO_SIGN = "This message attests that I control the given wallet"

def get_clob_auth_domain():
    return make_domain(name=CLOB_DOMAIN_NAME, version=CLOB_VERSION, chainId=1)

def sign_clob_auth_message(signer: Signer, timestamp: int)-> str:
    clob_auth_msg = ClobAuth(address=signer.address, timestamp=str(timestamp), message=MSG_TO_SIGN)
    auth_struct_hash = Web3.keccak(clob_auth_msg.signable_bytes(get_clob_auth_domain()))
    return signer.sign(auth_struct_hash)






# Access levels
L0 = 0
L1 = 1
L2 = 2


CREDENTIAL_CREATION_WARNING = """🚨🚨🚨
Your credentials CANNOT be recovered after they've been created. 
Be sure to store them safely!
🚨🚨🚨"""


L1_AUTH_UNAVAILABLE = "A private key is needed to interact with this endpoint!"

L2_AUTH_UNAVAILABLE = "API Credentials are needed to interact with this endpoint!"




TIME = "/time"
CREATE_API_KEY = "/create-api-key"
GET_API_KEYS = "/get-api-keys"
TRADE_HISTORY = "/trade-history"
GET_ORDER = "/order/"
OPEN_ORDERS = "/open-orders"
POST_ORDER = "/order"
CANCEL = "/order"
CANCEL_ALL = "/cancel-all"



from dataclasses import dataclass
from typing import Any


@dataclass
class ApiCreds:
    api_key: str
    api_secret: str
    api_passphrase: str

@dataclass
class RequestArgs:
    method: str
    request_path: str
    body: Any = None

@dataclass
class LimitOrderArgs:
    price: float
    size: float
    side: str
    token_id: str

@dataclass
class MarketOrderArgs:
    size: float
    side: str
    token_id: str






import hmac
import hashlib
import base64


def build_hmac_signature(secret: str, timestamp: str, method: str, requestPath: str, body=None):
    """
    Creates an HMAC signature by signing a payload with the secret
    """
    base64_secret = base64.urlsafe_b64decode(secret)
    message = str(timestamp) + str(method) + str(requestPath)
    if body:
        # NOTE: Necessary to replace single quotes with double quotes
        # to generate the same hmac message as go and typescript
        message += str(body).replace("'", '"') 
    print("Message:")
    print(message)
    
    h = hmac.new(base64_secret, bytes(message, "utf-8"), hashlib.sha256)

    # ensure base64 encoded
    return base64.urlsafe_b64encode(h.digest())





import logging

from .orders.builder import OrderBuilder

from .headers import create_level_1_headers, create_level_2_headers
from .signer import Signer

from .endpoints import CANCEL, CANCEL_ALL, CREATE_API_KEY, GET_API_KEYS, GET_ORDER, OPEN_ORDERS, POST_ORDER, TIME, TRADE_HISTORY
from .clob_types import ApiCreds, LimitOrderArgs, MarketOrderArgs, RequestArgs
from .exceptions import PolyException
from .http_helpers.helpers import delete, get, post
from .constants import CREDENTIAL_CREATION_WARNING, L0, L1, L1_AUTH_UNAVAILABLE, L2, L2_AUTH_UNAVAILABLE


class ClobClient:
    def __init__(self, host, chain_id: int = None, key:str = None, creds:ApiCreds = None, signature_type: int = None, funder: str = None):
        """
        Initializes the clob client
        The client can be started in 3 modes:
        1) Level 0: Requires only the clob host url
                    Allows access to open CLOB endpoints
        
        2) Level 1: Requires the host, chain_id and a private key.
                    Allows access to L1 authenticated endpoints + all unauthenticated endpoints

        3) Level 2: Requires the host, chain_id, a private key, and Credentials.
                    Allows access to all endpoints
        """
        self.host = host
        self.signer = Signer(key, chain_id) if key else None
        self.creds = creds
        self.mode = self._get_client_mode()
        if self.signer:
            self.builder = OrderBuilder(self.signer, sig_type=signature_type, funder=funder) 
        self.logger = logging.getLogger(self.__class__.__name__)

    def get_address(self):
        """
        Returns the public address of the signer
        """
        return self.signer.address if self.signer else None
    
    def get_ok(self):
        """
        Health check: Confirms that the server is up
        Does not need authentication
        """
        return get("{}/".format(self.host))
    

    def get_server_time(self):
        """
        Returns the current timestamp on the server
        Does not need authentication
        """
        return get("{}{}".format(self.host, TIME))

    def create_api_key(self):
        """
        Creates a new CLOB API key for the given 
        """
        self.assert_level_1_auth()
        
        endpoint = "{}{}".format(self.host, CREATE_API_KEY)
        headers= create_level_1_headers(self.signer)
        
        creds = post(endpoint, headers=headers)
        self.logger.info(CREDENTIAL_CREATION_WARNING)
        self.logger.info(creds)
        return creds

    def get_api_keys(self):
        """
        Gets the available API keys for this address
        Level 2 Auth required
        """
        self.assert_level_2_auth()
        
        request_args = RequestArgs(method="GET", request_path=GET_API_KEYS)
        headers = create_level_2_headers(self.signer, self.creds, request_args)
        return get("{}{}".format(self.host, GET_API_KEYS), headers=headers)

    def create_limit_order(self, order_args: LimitOrderArgs):
        """
        Creates and signs a limit order
        Level 2 Auth required
        """
        self.assert_level_2_auth()

        return self.builder.create_limit_order(order_args)

    def create_market_order(self, order_args: MarketOrderArgs):
        """
        Creates and signs a market order
        Level 2 Auth required
        """
        self.assert_level_2_auth()
        return self.builder.create_market_order(order_args)

    def post_order(self, order):
        """
        Posts the order
        """
        self.assert_level_2_auth()
        body = order.dict()
        headers = create_level_2_headers(self.signer, self.creds, RequestArgs(method="POST", request_path=POST_ORDER, body=body))
        return post("{}{}".format(self.host, POST_ORDER), headers=headers, data=body)

    def cancel(self, order_id):
        """
        Cancels an order
        Level 2 Auth required
        """
        self.assert_level_2_auth()
        body = {
            "orderID": order_id
        }

        request_args = RequestArgs(method="DELETE", request_path=CANCEL, body=body)
        headers = create_level_2_headers(self.signer, self.creds, request_args)
        return delete("{}{}".format(self.host, CANCEL), headers=headers, body=body)

    def cancel_all(self):
        """
        Cancels all available orders for the user
        Level 2 Auth required
        """
        self.assert_level_2_auth()
        request_args = RequestArgs(method="DELETE", request_path=CANCEL_ALL)
        headers = create_level_2_headers(self.signer, self.creds, request_args)
        return delete("{}{}".format(self.host, CANCEL_ALL), headers=headers)

    def get_open_orders(self):
        """
        Gets open orders for the API key
        Requires Level 2 authentication
        """
        self.assert_level_2_auth()
        request_args = RequestArgs(method="GET", request_path=OPEN_ORDERS)
        headers = create_level_2_headers(self.signer, self.creds, request_args)
        return get("{}{}".format(self.host, OPEN_ORDERS), headers=headers)

    def get_order(self, order_id):
        """
        Fetches the order corresponding to the order_id
        Requires Level 2 authentication
        """
        self.assert_level_2_auth()
        endpoint = "{}{}".format(GET_ORDER, order_id)
        request_args = RequestArgs(method="GET", request_path=endpoint)
        headers = create_level_2_headers(self.signer, self.creds, request_args)
        return get("{}{}".format(self.host, endpoint), headers=headers)

    def get_trade_history(self):
        self.assert_level_2_auth()
        request_args = RequestArgs(method="GET", request_path=TRADE_HISTORY)
        headers = create_level_2_headers(self.signer, self.creds, request_args)
        return get("{}{}".format(self.host, TRADE_HISTORY), headers=headers)
    
    def assert_level_1_auth(self):
        """
        Level 1 Poly Auth
        """
        if self.mode < L1:
            raise PolyException(L1_AUTH_UNAVAILABLE)


    def assert_level_2_auth(self):
        """
        Level 2 Poly Auth
        """
        if self.mode < L2:
            raise PolyException(L2_AUTH_UNAVAILABLE)

    def _get_client_mode(self):
        if self.signer is not None and self.creds is not None:
            return L2
        if self.signer is not None:
            return L1
        return L0
            


